from moteur_avec_variables.proposition_avec_variables import *

class BaseConnaissances:
    """ Une base de connaissances destinée à contenir les propositions\
        et les règles d'un système de chaînage arrière.
    """

    def __init__(self, constructeur_de_regle):
        """ Construit une base de connaissances.

            Le paramètre ``constructeur_de_regle`` doit être une fonction prenant\
            deux arguments: la liste des conditions d'une règle et sa conclusion.
            La fonction doit retourner une règle du type désiré.

            :param contructeur_de_regle: une fonction construisant une règle.
        """

        self.faits = {}
        self.regles = {}
        self.constructeur_de_regle = constructeur_de_regle
        self.sym = 0

    def ajoute_un_fait(self, fait):
        """ Ajoute un fait dans la base de connaissances. 

            :param fait: un fait.
        """

        clef = tete(fait)
        if clef in self.faits:
            self.faits[clef].append(fait)
        else:
            self.faits[clef] = [fait]

    def ajoute_faits(self, faits):
        """ Ajoute une liste de faits dans la base de connaissances.

            :param list faits: une liste de faits.
        """

        for fait in faits:
            self.ajoute_un_fait(fait)

    def ajoute_une_regle(self, description):
        """ Ajoute une règle dans la base de connaissances étant donné sa\
            description.

            Une règle est décrite par une liste (ou un tuple) de deux\
            éléments : une liste de conditions et une conclusion.

            Les conditions et la conclusion doivent être des propositions.

            :param description: une description de règle.
        """

        regle = self.constructeur_de_regle(description)
        clef = tete(regle.conclusion)
        if clef in self.regles:
            self.regles[clef].append(regle)
        else:
            self.regles[clef] = [regle]

    def ajoute_regles(self, descriptions):
        """ Ajoute des règles dans la base de connaissances.

            L'argument est une liste de descriptions, chacune composée d'une\
            liste de conditions et d'une conséquence.

            :param list descriptions: une liste de descriptions de règles.
        """

        for description in descriptions:
            self.ajoute_une_regle(description)

    def choisir_faits_interessants(self, pattern):
        """ Retourne une liste de faits potentiellement pertinents pour tenter\ 
            l'unification d'un pattern.
        """

        clef = tete(pattern)
        faits = self.faits.get(clef, [])
        return faits

    def choisir_regles_interessantes(self, pattern, unificateur):
        """ Retourne une liste de règles potentiellement pertinentes pour tenter\
            l'unification d'un pattern.
        """

        regles_interessantes = []
        clef = tete(pattern)
        if clef in self.regles:
            anciennes_regles = self.regles[clef]
            for regle in anciennes_regles:
                regles_interessantes.append(self.nouvelle_instance(regle, unificateur))
        return regles_interessantes

    def nouvelle_instance(self, regle, unificateur):
        """ Crée une nouvelle instance d'une règle avec des variables uniques. """

        env = {}
        variables = set()
        for cond in regle.conditions:
            variables.update(lister_variables(cond))
        variables.update(lister_variables(regle.conclusion))
        for var in variables:
            self.sym = self.sym + 1
            nouvelle_var = '?{}'.format(self.sym)
            env[var] = nouvelle_var

        conditions = [unificateur.substitue(cond, env) for cond in regle.conditions]
        conclusion = unificateur.substitue(regle.conclusion, env)

        return self.constructeur_de_regle((conditions, conclusion))